#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "glFile.h"

modelData * loadModel( const char *fileName, const char *textureName )
{
   modelData *model = NULL;
   FILE *hFile = fopen( fileName, "rb" );

   /* check the file open command */
   if( hFile != NULL )
   {
      int fileLen;
      char *buffer;
      modelHeader *mdh;
      frame *frm;
      coord *pointList;
      mesh *triIndex, *bufIndexPtr;
      stTexCoord *st;
      stTexture *stPtr;
      int i;

      /* find length of file */
      fseek( hFile, 0, SEEK_END );
      fileLen = ftell( hFile );
      fseek( hFile, 0, SEEK_SET );

      /* read in file */
      buffer = malloc( fileLen+1 );
      fread( buffer, sizeof( char ), fileLen, hFile );

      /* get pointer to file header */
      mdh = (modelHeader *)buffer;

      /* construct the model data structure */
      model = (modelData *)malloc( sizeof( modelData ));

      /* create a pointlist */
      pointList = (coord *)malloc( sizeof( coord ) * mdh->numXYZ );
      model->pointList = pointList;
      model->numPoints = mdh->numXYZ;

      /* offset to points in first frame */
      frm = (frame *)&buffer[mdh->offsetFrames];

      for( i=0; i<mdh->numXYZ; i++ )
      {
         pointList[i].pt[0] = frm->scale[0] * frm->fp[i].v[0] + frm->translate[0];
         pointList[i].pt[1] = frm->scale[1] * frm->fp[i].v[1] + frm->translate[1];
         pointList[i].pt[2] = frm->scale[2] * frm->fp[i].v[2] + frm->translate[2];
      }

      /* see if there is a texture associated with the file */
      model->texWidth = 1;
      model->texHeight = 1;
      do {
         if( mdh->numSkins )
         {
            /* if skinname then use it otherwise use commandline texture */
            textureName = &buffer[mdh->offsetSkins];
         }

         if( textureName )
         {
            FILE *hTexFile = fopen( textureName, "rb" );

            /* check the file open command */
            if( hTexFile != NULL )
            {
               int imgWidth, imgHeight, texFileLen, imgBufferPtr, i;
               pcxHeader *pcxPtr;
               unsigned char *imgBuffer, *texBuffer, *pcxBufferPtr, *paletteBuffer;

               /* find length of file */
               fseek( hTexFile, 0, SEEK_END );
               texFileLen = ftell( hTexFile );
               fseek( hTexFile, 0, SEEK_SET );

               /* read in file */
               texBuffer = malloc( texFileLen+1 );
               fread( texBuffer, sizeof( char ), texFileLen, hTexFile );

               /* get the image dimensions */
               pcxPtr = (pcxHeader *)texBuffer;
               imgWidth = pcxPtr->size[0] - pcxPtr->offset[0] + 1;
               imgHeight = pcxPtr->size[1] - pcxPtr->offset[1] + 1;

               /* image starts at 128 from the beginning of the buffer */
               imgBuffer = malloc( imgWidth * imgHeight );
               imgBufferPtr = 0;
               pcxBufferPtr = &texBuffer[128];
               /* decode the pcx image */
               while( imgBufferPtr < (imgWidth * imgHeight) )
               {
                  if( *pcxBufferPtr > 0xbf )
                  {
                     int repeat = *pcxBufferPtr++ & 0x3f;
                     for( i=0; i<repeat; i++ )
                        imgBuffer[imgBufferPtr++] = *pcxBufferPtr;
                  } else {
                     imgBuffer[imgBufferPtr++] = *pcxBufferPtr;
                  }
                  pcxBufferPtr++;
               }
               /* read in the image palette */
               paletteBuffer = malloc( 768 );
               for( i=0; i<768; i++ )
                  paletteBuffer[i] = texBuffer[ texFileLen-768+i ];

               /* find the nearest greater power of 2 for each dimension */
               {
                  int imageWidth = imgWidth, imageHeight = imgHeight;
                  i = 0;
                  while( imageWidth )
                  {
                     imageWidth /= 2;
                     i++;
                  }
                  model->texWidth = pow( 2, i );
                  i = 0;
                  while( imageHeight )
                  {
                     imageHeight /= 2;
                     i++;
                  }
                  model->texHeight = pow( 2, i );
               }
               /* now create the OpenGL texture */
               {
                  int i, j;
                  model->texture = malloc( model->texWidth * model->texHeight * 3 );
                  for (j = 0; j < imgHeight; j++)
                  {
                     for (i = 0; i < imgWidth; i++)
                     {
                        model->texture[3*(j * model->texWidth + i)+0]
                              = paletteBuffer[ 3*imgBuffer[j*imgWidth+i]+0 ];
                        model->texture[3*(j * model->texWidth + i)+1]
                              = paletteBuffer[ 3*imgBuffer[j*imgWidth+i]+1 ];
                        model->texture[3*(j * model->texWidth + i)+2]
                              = paletteBuffer[ 3*imgBuffer[j*imgWidth+i]+2 ];
                     }
                  }
               }
               /* cleanup */
               free( paletteBuffer );
               free( imgBuffer );
            } else {
               /* skip the texture setup functions */
               break;
            }
         } else {
            /* skip the texture setup functions */
            break;
         }

         /* create the texture list */
         st = (stTexCoord *)malloc( sizeof( stTexCoord ) * mdh->numST );
         model->numST = mdh->numST;
         model->st = st;

         stPtr = (stTexture *)&buffer[mdh->offsetST];
         for( i=0; i<mdh->numST; i++ )
         {
            st[i].s = (float)stPtr[i].s / (float)model->texWidth;
            st[i].t = (float)stPtr[i].t / (float)model->texWidth;
         }

      } while( 0 );                      /* only do once */

      /* create a mesh list */
      triIndex = (mesh *)malloc( sizeof( mesh ) * mdh->numTris );
      model->numTriangles = mdh->numTris;
      model->triIndex = triIndex;

      /* point to indexes in buffers */
      bufIndexPtr = (mesh *)&buffer[mdh->offsetTris];

      for( i=0; i<mdh->numTris; i++ )
      {
         triIndex[i].meshIndex[0] = bufIndexPtr[i].meshIndex[0];
         triIndex[i].meshIndex[1] = bufIndexPtr[i].meshIndex[1];
         triIndex[i].meshIndex[2] = bufIndexPtr[i].meshIndex[2];
         triIndex[i].stIndex[0] = bufIndexPtr[i].stIndex[0];
         triIndex[i].stIndex[1] = bufIndexPtr[i].stIndex[1];
         triIndex[i].stIndex[2] = bufIndexPtr[i].stIndex[2];
      }

      /* cleanup */
      free( buffer );
   }
   return( model );
}

